/*

  $Header: /export/home/lord/src/tube/RCS/tube.c,v 1.25 1996/01/12 20:15:21 lord Exp $

  $Log: tube.c,v $
  Revision 1.25  1996/01/12 20:15:21  lord
  Add remote client name/address loging.
  Remove SO_REUSEADDR.
  Some comments.
  TUBE_VERSION macro.

  Revision 1.24  1996/01/12 19:12:23  lord
  Some brush up with lint

  Revision 1.23  1996/01/12 01:00:59  lord
  Undef DEBUG

  Revision 1.22  1996/01/12 01:00:33  lord
  Remove debug printf.

  Revision 1.21  1996/01/12 00:47:46  lord
  Add some useless SO_REUSEADDR

  Revision 1.20  1996/01/11 23:51:21  lord
  Some experimenta with telnet echo.

  Revision 1.19  1996/01/11 23:43:01  lord
  Now dosassosiates from control term.

  Revision 1.18  1996/01/11 23:31:19  lord
  Works as concurrent server. Still not detach from tty.

  Revision 1.17  1996/01/11 22:25:53  lord
  Start redoing it to be run as usual daemon, not from inetd.

  Revision 1.16  1996/01/11 19:43:03  lord
  setpgrp(); call added.

  Revision 1.15  1996/01/11 18:49:00  lord
  Added:
  	int tty_raw(int fd);
  	int tty_reset(int fd);
  	int tty_getmode(int oldfd);
  	int tty_setmode(int newfd);

  Revision 1.14  1996/01/11 03:00:07  lord
  Debugging.

  Revision 1.13  1996/01/11 02:47:12  lord
  com_gw replaced with Stevens' pass_all()

  Revision 1.12  1996/01/11 02:17:16  lord
  Redo my_popen using ptys. Take some code from R.Stevens "Unix network programming".

  Revision 1.11  1996/01/11 01:03:01  lord
  Add some socket options

  Revision 1.10  1996/01/11 00:24:30  lord
  Add wait in pclose();

  Revision 1.9  1996/01/08 23:03:28  lord
  change fprintfs to syslog. closing stderr.

  Revision 1.8  1996/01/04 05:02:35  lord
  Add \E to send/exp

  Revision 1.7  1996/01/04 04:41:52  lord
  Send/Expect works!

  Revision 1.6  1996/01/04 03:29:16  lord
  Implement popen and pclose using socketpair

  Revision 1.5  1996/01/03 01:08:34  lord
  Add empty functions for my_popen and my_pclose

  Revision 1.4  1996/01/02 20:41:16  lord
  do_send,do_expect implemented.

  Revision 1.3  1996/01/02 19:59:04  lord
  preprocess_line() implemented.

  Revision 1.2  1995/12/30 01:36:01  lord
  *** empty log message ***

  Revision 1.1  1995/12/29 23:31:41  lord
  Initial revision

  */


/*
  Simple proxy daemon using telnet-like command as transport 
  ========================================================--
  Vadim Zaliva, lord@crocodile.org
  
  This software is provided ``as is'' without express or implied
  warranty.  Feel free to distribute it.  Part of this code are taken
  from W. RICHARD STEVENS "Unix Network Programming" excellent book.
  
  This code should be considered alpha.  Feel free to contact me with
  improvments requests and bug reports.
  
  */

/* #define DEBUG 1 */

#include <stdio.h>
#include "cfg.h"
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/filio.h>
#include <sys/stat.h>
#include <syslog.h>
#include <netdb.h>
#include <fcntl.h>
#include <stropts.h>
#include <termio.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define nil             NULL            /* my favourite :-) */
#define DEFAULT_TOUT    10
#define MBUFSIZ         8192 
#define MAX_CMD_WORDS   256
#define EXEC_DELAY      5
#define EXPECT_BUFFER_SIZE 256 /* This buffer shold be longer than any Send/Expect line */
#define	PTY_MASTER	"/dev/ptmx"	/* System V Release 3.2 */
#define DELAY_CHAR      '\007'
#define TUBE_VERSION    "tube v1.0a by lord@crocodile.org"

static int  send_expect(struct Dict *,int);
static void preprocess_line(struct Dict *,char *,char *,int *);
static int  do_send(char *,int,int);
static int  do_expect(char *,int,int);
static int  my_popen(char *,int *); 
static int  my_pclose(int);
static char **split_cmd_line(char *);
static int  pty_master(void);
static int  pty_slave(int);
static void pass_all(int , int);
static void sig_term(int);
static int  writen(int , char *, int );
static void daemon_start(void);
static void exec_shell(int , char **);
 
static int  sigcaught;	/* flag set by signal handler */

int main(int ac, char **av)
{
    struct Dict *tubecfg=NULL;
    int  telnet;
    char *tmp;
    char telnet_cmd[MAXPATHLEN];
    int childpid;
    int sockfd,newsockfd;
    int clien,childpid1;
    struct sockaddr_in cli_addr,serv_addr;
    int portn;
    struct hostent *hp;
    char *remote_name;
    
    openlog("tube", LOG_PID| LOG_CONS|LOG_NOWAIT, LOG_USER);
    syslog(LOG_INFO,"Starting %s",TUBE_VERSION);

    if(ac!=2) {
	syslog(LOG_ERR,"USAGE: %s <cfgfile>\n",av[0]);
	closelog();
	exit(1);
    }
    
    switch(readcfg(av[1],&tubecfg))
    {
    case CFG_BADCFG:
    {
	syslog(LOG_ERR,"Can't open file: %s\n",av[1]);
	closelog();
	exit(1);
    }
    case CFG_EMPTYCFG:
    {
	syslog(LOG_ERR,"No records in %s file\n",av[1]);
	closelog();
	exit(1);
    }
    }
    
    if((tmp=cfgfind("Telnet",tubecfg))==NULL)
    {
	syslog(LOG_ERR,"No 'Telnet' parametr in config file %s\n",av[1]);
	freecfg(tubecfg);
	closelog();
	exit(1);
    }  else strcpy(telnet_cmd,tmp);

    if((tmp=cfgfind("Port",tubecfg))==NULL)
    {
	syslog(LOG_ERR,"No 'Port' parametr in config file %s. Assuming 119\n",av[1]);
	portn=119;
    }  else portn=atoi(tmp);
    
    if(portn<=1024 && geteuid()!=0)
    {
	syslog(LOG_ERR,"You must be root to run TUBE on reserved port\n");
	freecfg(tubecfg);
	closelog();
	exit(1);
    }

    /* Let's become a daemon */

    daemon_start(); 

    if((sockfd=socket(AF_INET,SOCK_STREAM,0))<0)
    {
	syslog(LOG_ERR,"Error creating socket.");
	freecfg(tubecfg);
	closelog();
	exit(1);
    }

    memset((void *)&serv_addr,0,sizeof(serv_addr));
    serv_addr.sin_family=AF_INET;
    serv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
    serv_addr.sin_port=htons(portn);
    
    if(bind(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))<0)
    {
	syslog(LOG_ERR,"Error binding socket.");
	freecfg(tubecfg);
	closelog();
	exit(1);
    }
    
    syslog(LOG_INFO,"Waiting for connections.");
    if(listen(sockfd,5)<0)
    {
	syslog(LOG_ERR,"Error listening socket.");
	freecfg(tubecfg);
	closelog();
	exit(1);
    }

    while(1)
    {
	clien=sizeof(cli_addr);
	newsockfd=accept(sockfd,(struct sockaddr *)&cli_addr,&clien);

	if(newsockfd<0)
	{
	    syslog(LOG_ERR,"accept error");
	}

	if((childpid1=fork())<0)
	{
	    syslog(LOG_ERR,"error fork");
	}
	
	if(childpid1 == 0) /* Child */
	{
	    hp=gethostbyaddr((char *)&cli_addr,clien,AF_INET);
	    remote_name=strdup(hp?(hp->h_name):inet_ntoa(cli_addr.sin_addr));
	
	    syslog(LOG_INFO,"Got connection from: %s",remote_name); 
	    
	    close(sockfd);
	    /* Process connection */

	    if((telnet=my_popen(telnet_cmd,&childpid))<0)
	    {
		syslog(LOG_ERR,"Error executing 'Telnet'.\n");
		freecfg(tubecfg);
		closelog();
		exit(1);
	    }

	    /* We get here only in parent */
        
	    if(send_expect(tubecfg,telnet))
	    {
		syslog(LOG_ERR,"Send-expect failed\n");
		my_pclose(telnet);
		freecfg(tubecfg);
		closelog();
		exit(1);
	    }
	    
	    pass_all(telnet,newsockfd);
	    
	    syslog(LOG_INFO,"Client %s done.\n",remote_name);
	    free(remote_name);
	    (void)my_pclose(telnet);
	    exit(0);
	}
	close (newsockfd);
    }
 
    /* NOTREACHED */
    freecfg(tubecfg);
    closelog();
    return(0);
}


static int send_expect(struct Dict *tubecfg,int telnet)
{
    int  i;
    char cmd_name[80];
    char expect[1024];
    char send[1024];
    int  tout;
    char *tmp;
    
    i=1;
    while(1)
    {
	tout=DEFAULT_TOUT;
	sprintf(cmd_name,"Expect-%d",i);
	if((tmp=cfgfind(cmd_name,tubecfg))==NULL)
	    break;
	else
	    preprocess_line(tubecfg,tmp,expect,&tout);
	
#ifdef DEBUG
	syslog(LOG_DEBUG,"EXPECT: >%s<\n",tmp);
#endif
	
	if(do_expect(expect,telnet,tout))
	{
	    syslog(LOG_ERR,"Expect failed!\n");
	    return 1;	
	}
	
	sprintf(cmd_name,"Send-%d",i);
	if((tmp=cfgfind(cmd_name,tubecfg))==NULL)
	    break;
	else
	    preprocess_line(tubecfg,tmp,send,&tout);
	
	i++;

#ifdef DEBUG
	syslog(LOG_DEBUG,"SEND  : %s\n",tmp);
#endif
	
	if(do_send(send,telnet,tout))
	{
	    syslog(LOG_ERR,"Send failed!\n");
	    return 1;	
	}
	
	
    }

    return 0;
}

static void preprocess_line(struct Dict *tubecfg,char *from,char *to,int *tout)
{
    char *f;
    char *t;
    char *tmp;
    
/*   \E - ^E                            */
/*   \U - user                          */
/*   \P - password                      */
/*   \n - newline  (0x0a)               */
/*   \d - 2 sec. delay                  */
/*   \r - linefeed (0x0d)               */
/*   \TXXX - set tout to XXX seconds */
    
    f=from;t=to;

    while(*f)
    {
	if(*f=='\\')
	{
	    f++;
	    switch(*f) {
		
	    case 'U':
	    {
		if((tmp=cfgfind("User",tubecfg))==NULL)
		    tmp="USER";
		strcpy(t,tmp);
		t=strchr(t,'\0');
		f++;
		break;
	    }

	    case 'P':
	    {
		if((tmp=cfgfind("Password",tubecfg))==NULL)
		    tmp="USER";
		strcpy(t,tmp);
		t=strchr(t,'\0');
		f++;
		break;
	    }
		
	    case 'n':
	    {
		*t='\n';
		t++;
		f++;
		break;
	    }

	    case 'E':
	    {
		*t=0x05;
		t++;
		f++;
		break;
	    }

	    case 'D':
	    {
		*t=DELAY_CHAR;
		t++;
		f++;
		break;
	    }
	    
	    case 'r':
	    {
		*t='\r';
		t++;
		f++;
		break;
	    }
	    
	    case 'T':
	    {
		*tout=*(f+1)*100+*(f+2)*10+*(f+3);
		f+=4;
	    }

	    default: 
	    {
		*t=*f;
		t++;
		f++;
		break;
	    }
	    
	    } /* switch */
	} else {
	    *t=*f;
	    t++;
	    f++;
	}
    } /* while */

    *t='\0';
}

static int do_send(char *str,int f,int tout)
{
    time_t started;
    
    started=time(nil);
    while(*str)
    {
	if((time(nil)-started) > tout)
	    return 1;
	
	if(*str==DELAY_CHAR)
	{
	    sleep(2);
	} else {
#ifdef DEBUG
	    syslog(LOG_DEBUG,"PUT: %c\n",*str);
#endif
	    while(write(f,str,1)==-1)
	    {
#ifdef DEBUG
		syslog(LOG_DEBUG,"REPUT: %c\n",*str);
#endif
		if((time(nil)-started) > tout) 
		    return 1;
	    }
	}
	
	str++;
    }
    return 0;
}

static int do_expect(char *str,int f,int tout)
{
    time_t started;
    char   buf[EXPECT_BUFFER_SIZE+2];
    int    i,j;
    char   *endstr;
    char   *tmp;
    
    while(*str && *str==DELAY_CHAR)
    {
	sleep(2);
	str++;
    }

    if(!*str) return 0; /* Empty line always match */

    endstr=strchr(str,'\0')-1;
    memset(buf,0,sizeof(buf));
    i=0;
    started=time(nil);
    while((time(nil)-started) <= tout)
    {
	if(read(f,buf+i,1)==1) /* TODO: poll */
	{
#ifdef DEBUG	    
	    syslog(LOG_DEBUG,"GOT: %c\n",buf[i]);
#endif	    
	    tmp=endstr;
	    j=i;
	    while(tmp!=str && *tmp==buf[j])
	    {
		tmp--;
		j--;
		if(j==-1) j=EXPECT_BUFFER_SIZE;

	    }
	    if(tmp==str && *tmp==buf[j]) return 0; /* Match */
	    
	    i++;
	    if(i==EXPECT_BUFFER_SIZE+1) i=0;
	}
    }
    
#ifdef DEBUG
    syslog(LOG_DEBUG,"Expect timeout\n");
#endif		
    return 1;
}

/* TODO: env */
static int my_popen(char *cmd,int *r_child)
{
    int	master_fd, slave_fd, childpid;

    if ( (master_fd = pty_master()) < 0)
    {
	syslog(LOG_ERR,"can't open master pty");
	return -1;
    }

    if ((childpid = fork()) < 0)
    {
	syslog(LOG_ERR,"can't fork");
    }
    else if(childpid == 0) 	/* child process */
    {
	/*
	 * Disassociate from our control terminal so we can acquire
	 * the pseudo-terminal slave as our control tty.
	 * This call also makes us a process group leader, which is
	 * necessary to acquire a control tty.
	 */
	
	setpgrp();
	
	if((slave_fd = pty_slave(master_fd)) < 0)
	{
	    syslog(LOG_ERR,"can't open pty slave");
	    return -1;
	}
	close(master_fd);

	exec_shell(slave_fd, split_cmd_line(cmd));
	/* NOTREACHED */
    }

    *r_child=childpid;
    return(master_fd);
}

static int my_pclose(int fd)
{
    return close(fd);
}

static char ** split_cmd_line(char *cmd)
{
    int i;
    static char *a[MAX_CMD_WORDS];
    char *tmp;

    i=0;
    tmp=strtok(cmd," \t");
    while(tmp && (i< MAX_CMD_WORDS))
    {
	a[i]=strdup(tmp);
	i++;
	tmp=strtok(nil," \t");
    }
    a[i]=NULL;
    return a;
}

/*
 * Opens master half of pseudo-terminal
 * returns the file descriptor, or -1 on error
 */
static int  pty_master(void)
{
    int	master_fd;

    /*
     * Open the master half - "/dev/ptms".  This is a streams clone
     * device, so it'll allocate the first available pty master.
     */

    if ( (master_fd = open(PTY_MASTER, O_RDWR)) < 0)
	return(-1);

    return(master_fd);
}

/*
 * Open the slave half of a pseudo-terminal.
 * returns the file descriptor, or -1 on error 
 */
static int pty_slave(int master_fd)
{
    int	slave_fd;
    char	*slavename;
    int	 grantpt (int);	/* undocumented function */
    int	 unlockpt(int);	/* undocumented function */
    char *ptsname(int);	/* undocumented function */

    if (grantpt(master_fd) < 0) {	/* change permissions of slave */
	close(master_fd);
	return(-1);
    }
    
    if (unlockpt(master_fd) < 0) {	/* unlock slave */
	close(master_fd);
	return(-1);
    }

    slavename = ptsname(master_fd);	/* determine the slave's name */
    if (slavename == NULL) {
	close(master_fd);
	return(-1);
    }

    slave_fd = open(slavename, O_RDWR);	/* open the slave */
    if (slave_fd < 0) {
	close(master_fd);
	return(-1);
    }

    /*
     * Now push two modules onto the slave's stream: "ptem" is the
     * pseudo-terminal hardware emulation module, and "ldterm" is
     * the standard terminal line discipline.
     */

    if (ioctl(slave_fd, I_PUSH, "ptem") < 0) {
	close(master_fd);
	return(-1);
    }
    if (ioctl(slave_fd, I_PUSH, "ldterm") < 0) {
	close(master_fd);
	return(-1);
    }

    return(slave_fd);
}

static void exec_shell(int fd, char **argv)
{
    close(0); close(1); close(2);
    if (dup(fd) != 0 || dup(fd) != 1 || dup(fd) != 2)
	syslog(LOG_ERR,"dup error");
    close(fd);
    
    execv(argv[0], argv);
    syslog(LOG_ERR,"execv error");
    /* NOTREACHED */
}

/*
 * If we get here, the child that was copying stdin to the IPC
 * channel got an EOF or error.  It has notified us with a SIGTERM
 * signal.  We set a flag for the parent process above.
 */
static void sig_term(int notused)
{
    sigcaught = 1;		/* set flag */
}

/*
 * Write "n" bytes to a descriptor.
 * Use in place of write() when fd is a stream socket.
 */
static int writen(int fd, char *ptr, int nbytes)
{
    int	nleft, nwritten;

    nleft = nbytes;
    while (nleft > 0) {
	nwritten = write(fd, ptr, nleft);
	if (nwritten <= 0)
	    return(nwritten);		/* error */

	nleft -= nwritten;
	ptr   += nwritten;
    }
    return(nbytes - nleft);
}

/*
 * Detach a daemon process from login session context.
 */
static void daemon_start(void)
{

    /* Maybe I should do 2 forks here? */
    
    if (fork())
	exit(0);
    chdir("/");
    umask(0);
    (void) close(0);
    (void) close(1);
    (void) close(2);
    (void) open("/", O_RDONLY);
    (void) dup2(0, 1);
    (void) dup2(0, 2);
    setsid();
}

static void pass_all(int fd, int client)
{
    int		newpid, nread;
    char	buff[MBUFSIZ];
    
    if ( (newpid = fork()) < 0) {
	syslog(LOG_ERR,"parent1: can't fork");
    } else
	if (newpid == 0)
	{	/* child: client -> fd */
	    while(1)
	    {
		nread = read(client, buff, MBUFSIZ);
		if (nread < 0)
		    syslog(LOG_ERR,"read error from stdin");
		else if (nread == 0)
		    break;		/* stdin EOF -> done */
		
		if (writen(fd, buff, nread) != nread)
		    syslog(LOG_ERR,"writen error to stream pipe");
	    }
	    kill(getppid(), SIGTERM);	/* kill parent */
	    exit(0);
	}
    /* parent: fd -> stdout */
    
    sigcaught = 0;
    signal(SIGTERM, sig_term);
    
    while (1)
    {
	if ( (nread = read(fd, buff, MBUFSIZ)) <= 0)
	    break;		/* error, EOF or signal; terminate */
	
	if (write(client, buff, nread) != nread)
	    syslog(LOG_ERR,"write error to client");
    }
    
    /*
     * If we get here either there was an EOF on the stream pipe,
     * implying that the shell terminated, or the child process above
     * terminated and we received its SIGTERM.
     * If the shell terminated, we have to let the child process
     * that we fork'ed above know this, so that it can break out
     * of its read from the standard input. 
     * If we received the signal, the child is already gone, so we're done.
     */
    
    if (sigcaught == 0)
	kill(newpid, SIGTERM);
    
    return;		/* parent returns to caller */
}


